from io import TextIOWrapper
import click
import os


DEFAULT_COLLECTION = "opensource_movies"
"""
开放 collection ，任何人均可上传。
通过 biliarchiver 上传的 item 会在24小时内被自动转移到 bilibili_videos collection
"""
BILIBILI_VIDEOS_COLLECTION = "bilibili_videos"
""" 由 arkiver 管理。bilibili_videos 属于 social-media-video 的子集 """
BILIBILI_VIDEOS_SUB_1_COLLECTION = "bilibili_videos_sub_1"
""" 由 yzqzss 管理。属于 bilibili_videos 的子集 """


@click.command(help=click.style("上传至互联网档案馆", fg="cyan"))
@click.option("--bvids", type=click.File(), default=None, help="bvids 列表的文件路径")
@click.option(
    "--by-storage-home-dir",
    is_flag=True,
    default=False,
    help="使用 `$storage_home_dir/videos` 目录下的所有视频",
)
@click.option("--update-existing", is_flag=True, default=False, help="更新已存在的 item")
@click.option(
    "--collection",
    default=DEFAULT_COLLECTION,
    type=click.Choice(
        [
            DEFAULT_COLLECTION,
            BILIBILI_VIDEOS_COLLECTION,
            BILIBILI_VIDEOS_SUB_1_COLLECTION,
        ]
    ),
    help=f"Collection to upload to. (非默认值仅限 collection 管理员使用) [default: {DEFAULT_COLLECTION}]",
)
def up(
    bvids: TextIOWrapper,
    by_storage_home_dir: bool,
    update_existing: bool,
    collection: str,
):
    from biliarchiver._biliarchiver_upload_bvid import upload_bvid
    from biliarchiver.config import config

    if by_storage_home_dir:
        for bvid_with_upper_part in os.listdir(config.storage_home_dir / "videos"):
            bvid = bvid_with_upper_part
            if "-" in bvid_with_upper_part:
                bvid = bvid_with_upper_part.split("-")[0]

            upload_bvid(bvid, update_existing=update_existing, collection=collection)

    elif bvids:
        bvids_from_file = bvids.read().splitlines()
        for bvid in bvids_from_file:
            upload_bvid(bvid, update_existing=update_existing, collection=collection)



import click
from rich.console import Console


@click.command(help=click.style("从哔哩哔哩下载", fg="cyan"))
@click.option("--bvids", type=click.Path(exists=False), required=True, help="bvids 列表的文件路径")
@click.option(
    "--skip-ia-check",
    "-s",
    is_flag=True,
    default=False,
    show_default=True,
    help="不检查 IA 上是否已存在对应 BVID 的 item ，直接开始下载",
)
@click.option(
    "--from-browser",
    "--fb",
    type=str,
    default=None,
    help="从指定浏览器导入 cookies (否则导入 config.json 中的 cookies_file)",
)
@click.option(
    "--min-free-space-gb",
    type=int,
    default=10,
    help="最小剩余空间 (GB)，用超退出",
    show_default=True,
)
@click.option("--skip-to", type=int, default=0, show_default=True, help="跳过文件开头 bvid 的个数")
def down(**kwargs):
    from biliarchiver.cli_tools.bili_archive_bvids import _down

    try:
        _down(**kwargs)
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
    finally:
        # 显示终端光标
        console = Console()
        console.show_cursor()



class VideosBasePathNotFoundError(FileNotFoundError):
    def __init__(self, path: str):
        self.path = path

    def __str__(self):
        return f"Videos base path {self.path} not found"

class VideosNotFinishedDownloadError(FileNotFoundError):
    def __init__(self, path: str):
        self.path = path

    def __str__(self):
        return f"Videos not finished download: {self.path}"

class VersionOutdatedError(Exception):
    def __init__(self, version):
        self.version = version

    def __str__(self):
        return "Version outdated: %s" % self.version


import asyncio
import os
from pathlib import Path
import re

import aiofiles
import httpx
from bilix.download.utils import raise_api_error, req_retry
from bilix.exception import APIError

from bilix.sites.bilibili import api

from rich import print
import json

from bilix.sites.bilibili.downloader import DownloaderBilibili
from bilix.exception import APIResourceError
from biliarchiver.config import BILIBILI_IDENTIFIER_PERFIX
from biliarchiver.config import config
from biliarchiver.utils.identifier import human_readable_upper_part_map

@raise_api_error
async def new_get_subtitle_info(client: httpx.AsyncClient, bvid, cid):
    params = {'bvid': bvid, 'cid': cid}
    res = await req_retry(client, 'https://api.bilibili.com/x/player/v2', params=params)
    info = json.loads(res.text)
    if info['code'] == -400:
        raise APIError('未找到字幕信息', params)

    # 这里 monkey patch 一下把返回 lan_doc 改成返回 lan，这样生成的字幕文件名就是 语言代码 而不是 中文名 了
    # 例如
    # lan_doc: 中文（中国）
    # lang: zh-CN

  # return [[f'http:{i["subtitle_url"]}', i['lan_doc']] for i in info['data']['subtitle']['subtitles']]
    return [[f'http:{i["subtitle_url"]}', i['lan']] for i in info['data']['subtitle']['subtitles']]
api.get_subtitle_info = new_get_subtitle_info


@raise_api_error
async def new_get_video_info(client: httpx.AsyncClient, url: str):
    """ 
    monkey patch 一下，只使用 API 获取 video_info
    理由：
        - API 目前只支持一般的 AV/BV 视频，可以预防我们不小心下到了番剧/影视剧之类的版权内容
        - 如果缺 url 对应的分P，bilix 那边会报 AssertionError(f"没有找到分P: p{selected_page_num}，请检查输入")
        - 对于一些老视频， _get_video_info_from_html() 经常返回烦人且不稳定的 durl 资源。而 API 会更多请求到 dash 资源（虽然仍有少数视频只有 durl 资源）。
    """
    # print("using api")
    return await api._get_video_info_from_api(client, url)
api.get_video_info = new_get_video_info

async def _new_attach_dash_and_durl_from_api(client: httpx.AsyncClient, video_info: api.VideoInfo):
    params = {'cid': video_info.cid, 'bvid': video_info.bvid,
              'qn': 120,  # 如无 dash 资源（少数老视频），fallback 到 4K 超清 durl
              'fnval': 4048,  # 如 dash 资源可用，请求 dash 格式的全部可用流
              'fourk': 1,  # 请求 4k 资源
              'fnver': 0, 'platform': 'pc', 'otype': 'json'}
    dash_response = await req_retry(client, 'https://api.bilibili.com/x/player/playurl',
                                    params=params, follow_redirects=True)
    dash_json = json.loads(dash_response.text)
    if dash_json['code'] != 0:
        raise APIResourceError(dash_json['message'], video_info.bvid)
    dash, other = None, []
    if 'dash' in dash_json['data']:
        dash = api.Dash.from_dict(dash_json)
    if 'durl' in dash_json['data']:
        for i in dash_json['data']['durl']:
            suffix = re.search(r'\.([a-zA-Z0-9]+)\?', i['url']).group(1) # type: ignore
            other.append(api.Media(base_url=i['url'], backup_url=i['backup_url'], size=i['size'], suffix=suffix))
    video_info.dash, video_info.other = dash, other
# NOTE: 临时修复，等 bilix 发布了 https://github.com/HFrost0/bilix/pull/174 的版本后，删掉这个 patch 
api._attach_dash_and_durl_from_api = _new_attach_dash_and_durl_from_api


async def archive_bvid(d: DownloaderBilibili, bvid: str, *, logined: bool=False, semaphore: asyncio.Semaphore):
    async with semaphore:
        assert d.hierarchy is True, 'hierarchy 必须为 True' # 为保持后续目录结构、文件命名的一致性
        assert d.client.cookies.get('SESSDATA') is not None, 'sess_data 不能为空' # 开个大会员呗，能下 4k 呢。
        assert logined is True, '请先检查 SESSDATA 是否过期，再将 logined 设置为 True' # 防误操作
        upper_part = human_readable_upper_part_map(string=bvid, backward=True)
        OLD_videos_basepath: Path = config.storage_home_dir / 'videos' / bvid
        videos_basepath: Path = config.storage_home_dir / 'videos' / f'{bvid}-{upper_part}'

        if os.path.exists(OLD_videos_basepath):
            print(f'检测到旧的视频目录 {OLD_videos_basepath}，将其重命名为 {videos_basepath}...')
            os.rename(OLD_videos_basepath, videos_basepath)


        if os.path.exists(videos_basepath / '_all_downloaded.mark'):
            print(f'{bvid} 所有分p都已下载过了')
            return

        url = f'https://www.bilibili.com/video/{bvid}/'
        # 为了获取 pages，先请求一次
        try:
            first_video_info = await api.get_video_info(d.client, url)
        except APIResourceError as e:
            print(f'{bvid} 获取 video_info 失败，原因：{e}')
            return

        os.makedirs(videos_basepath, exist_ok=True)

        pid = 0
        for page in first_video_info.pages:
            pid += 1 # pid 从 1 开始
            if not page.p_url.endswith(f'?p={pid}'):
                raise NotImplementedError(f'{bvid} 的 P{pid} 不存在 (可能视频被 UP主/B站 删了)，请报告此问题，我们需要这个样本！')

            file_basename = f'{bvid}_p{pid}'
            video_basepath = videos_basepath / f'{BILIBILI_IDENTIFIER_PERFIX}-{file_basename}'
            video_extrapath = video_basepath / 'extra'
            if os.path.exists(f'{video_basepath}/_downloaded.mark'):
                print(f'{file_basename}: 已经下载过了')
                continue

            def delete_cache(reason: str = ''):
                if not os.path.exists(video_basepath):
                    return
                _files_in_video_basepath = os.listdir(video_basepath)
                for _file in _files_in_video_basepath:
                    if _file.startswith(file_basename):
                        print(f'{file_basename}: {reason}，删除缓存: {_file}')
                        os.remove(video_basepath / _file)
            delete_cache('为防出错，清空上次未完成的下载缓存')
            video_info = await api.get_video_info(d.client, page.p_url)
            print(f'{file_basename}: {video_info.title}...')
            os.makedirs(video_basepath, exist_ok=True)
            os.makedirs(video_extrapath, exist_ok=True)


            old_p_name = video_info.pages[video_info.p].p_name
            old_h1_title = video_info.h1_title
        
            # 在 d.hierarchy is True 且 h1_title 超长的情况下， bilix 会将 p_name 作为文件名
            video_info.pages[video_info.p].p_name = file_basename # 所以这里覆盖 p_name 为 file_basename
            video_info.h1_title = 'iiiiii' * 50 # 然后假装超长标题
            # 这样 bilix 保存的文件名就是我们想要的了（谁叫 bilix 不支持自定义文件名呢）
            # NOTE: p_name 似乎也不宜过长，否则还是会被 bilix 截断。
            # 但是我们以 {bvid}_p{pid} 作为文件名，这个长度是没问题的。


            codec = None
            quality = None
            if video_info.dash:
                # 选择编码 dvh->hev->avc
                # 不选 av0 ，毕竟目前没几个设备能拖得动
                codec_candidates = ['dvh', 'hev', 'avc']
                for codec_candidate in codec_candidates:
                    for media in video_info.dash.videos:
                        if media.codec.startswith(codec_candidate):
                            codec = media.codec
                            quality = media.quality
                            print(f'{file_basename}: "{codec}" "{media.quality}" ...')
                            break
                    if codec is not None:
                        break
                assert codec is not None and quality is not None, f'{file_basename}: 没有 dvh、avc 或 hevc 编码的视频'
            elif video_info.other:
                print(f'{file_basename}: 未解析到 dash 资源，交给 bilix 处理 ...')
                codec = ''
                quality = 0
            else:
                raise APIError(f'{file_basename}: 未解析到视频资源', page.p_url)

            assert codec is not None
            assert isinstance(quality, (int, str))

            cor1 = d.get_video(page.p_url ,video_info=video_info, path=video_basepath,
                        quality=quality, # 选择最高画质
                        codec=codec, # 编码
                        # 下载 ass 弹幕(bilix 会自动调用 danmukuC 将 pb 弹幕转为 ass)、封面、字幕
                        # 弹幕、封面、字幕都会被放进 extra 子目录里，所以需要 d.hierarchy is True
                        dm=True, image=True, subtitle=True
                        )
            # 下载原始的 pb 弹幕
            cor2 = d.get_dm(page.p_url, video_info=video_info, path=video_extrapath)
            # 下载视频超详细信息（BV 级别，不是分 P 级别）
            cor3 = download_bilibili_video_detail(d.client, bvid, f'{video_extrapath}/{file_basename}.info.json')
            coroutines = [cor1, cor2, cor3]
            tasks = [asyncio.create_task(cor) for cor in coroutines]
            results = await asyncio.gather(*tasks, return_exceptions=True)
            for result, cor in zip(results, coroutines):
                if isinstance(result, Exception):
                    print("出错，其他任务完成后将抛出异常...")
                    for task in tasks:
                        task.cancel()
                    raise result

            if codec.startswith('hev') and not os.path.exists(video_basepath / f'{file_basename}.mp4'):

                # 如果有下载缓存文件（以 file_basename 开头的文件），说明这个 hevc 的 dash 资源存在，只是可能因为网络之类的原因下载中途失败了
                delete_cache('下载出错')

                # 下载缓存文件都不存在，应该是对应的 dash 资源根本就没有，一些老视频会出现这种情况。
                # 换 avc 编码
                print(f'{file_basename}: 视频文件没有被下载？也许是 hevc 对应的 dash 资源不存在，尝试 avc ……')
                assert video_info.dash is not None
                for media in video_info.dash.videos:
                    if media.codec.startswith('avc'):
                        codec = media.codec
                        print(f'{file_basename}: "{codec}" "{media.quality}" ...')
                        break
                cor4 = d.get_video(page.p_url ,video_info=video_info, path=video_basepath,
                        quality=0, # 选择最高画质
                        codec=codec, # 编码
                        # 下载 ass 弹幕(bilix 会自动调用 danmukuC 将 pb 弹幕转为 ass)、封面、字幕
                        # 弹幕、封面、字幕都会被放进 extra 子目录里，所以需要 d.hierarchy is True
                        dm=True, image=True, subtitle=True
                        )
                await asyncio.gather(cor4)


            assert os.path.exists(video_basepath / f'{file_basename}.mp4') or os.path.exists(video_basepath / f'{file_basename}.flv')

            # 还原为了自定义文件名而做的覆盖
            video_info.pages[video_info.p].p_name = old_p_name
            video_info.h1_title = old_h1_title

            # 单 p 下好了
            async with aiofiles.open(f'{video_basepath}/_downloaded.mark', 'w', encoding='utf-8') as f:
                await f.write('')


        # bv 对应的全部 p 下好了
        async with aiofiles.open(f'{videos_basepath}/_all_downloaded.mark', 'w', encoding='utf-8') as f:
                await f.write('')

    


async def download_bilibili_video_detail(client, bvid, filename):
    if os.path.exists(filename):
        print(f'{bvid} 视频详情已存在')
        return
    # url = 'https://api.bilibili.com/x/web-interface/view'
    url = 'https://api.bilibili.com/x/web-interface/view/detail' # 超详细 API（BV 级别，不是分 P 级别）
    params = {'bvid': bvid}
    r = await req_retry(client, url, params=params ,follow_redirects=True)
    r.raise_for_status()
    r_json = r.json()
    assert r_json['code'] == 0, f'{bvid} 视频详情获取失败'

    async with aiofiles.open(filename, 'w', encoding='utf-8') as f:
        # f.write(json.dumps(r.json(), indent=4, ensure_ascii=False))
        await f.write(r.text)
    print(f'{bvid} 视频详情已保存')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['biliarchiver', 'biliarchiver.cli_tools', 'biliarchiver.utils']

package_data = \
{'': ['*']}

install_requires = \
['bilix==0.18.4',
 'browser-cookie3>=0.19.1,<0.20.0',
 'click-option-group>=0.5.6,<0.6.0',
 'click>=8.1.6,<9.0.0',
 'danmakuc>=0.3.6,<0.4.0',
 'internetarchive>=3.5.0,<4.0.0']

entry_points = \
{'console_scripts': ['biliarchiver = '
                     'biliarchiver.cli_tools.biliarchiver:biliarchiver']}

setup_kwargs = {
    'name': 'biliarchiver',
    'version': '0.1.0',
    'description': '',
    'long_description': '# biliarchiver\n\n> 基于 bilix 的 BiliBili 存档工具\n\n## Install\n\n```bash\npip install biliarchiver\n```\n\n## Usage\n\n待补充。\n',
    'author': 'yzqzss',
    'author_email': 'yzqzss@yandex.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import asyncio
import os
from pathlib import Path
import re
import time
from httpx import AsyncClient
import requests
import json
import click
from click_option_group import optgroup

from bilix.sites.bilibili import api
from rich import print


""" def arg_parse():
    parser = argparse.ArgumentParser()
    # 为啥是 by-xxx 而不是 from-xxx ？因为命令行里好敲……
    ranking_group = parser.add_argument_group()
    ranking_group.title = 'by ranking'
    ranking_group.description = '排行榜（全站榜，非个性推荐榜）'
    ranking_group.add_argument(
        '--by-ranking', action='store_true', help='从排行榜获取 bvids')
    ranking_group.add_argument('--ranking-rid', type=int, default=0,
                               help='目标排行 rid，0 为全站排行榜。rid 等于分区的 tid [default: 0]')

    up_videos_group = parser.add_argument_group()
    up_videos_group.title = 'by up videos'
    up_videos_group.description = 'up 主用户页投稿'
    up_videos_group.add_argument(
        '--by-up_videos', action='store_true', help='从 up 主用户页获取全部的投稿的 bvids')
    up_videos_group.add_argument(
        '--up_videos-mid', type=str, help='目标 up 主的 mid (也可以是用户页的 URL)')

    popular_precious_group = parser.add_argument_group()
    popular_precious_group.title = 'popular precious'
    popular_precious_group.description = '入站必刷，更新频率低'
    popular_precious_group.add_argument(
        '--by-popular_precious', action='store_true', help='从入站必刷获取 bvids', dest='by_popular_precious')

    popular_series_group = parser.add_argument_group()
    popular_series_group.title = 'popular series'
    popular_series_group.description = '每周必看，每周五晚18:00更新'
    popular_series_group.add_argument(
        '--by-popular_series', action='store_true', help='从每周必看获取 bvids', dest='by_popular_series')
    popular_series_group.add_argument(
        '--popular_series-number', type=int, default=1, help='获取第几期（周） [default: 1]')
    popular_series_group.add_argument(
        '--all-popular_series', action='store_true', help='自动获取全部的每周必看（增量）', dest='all_popular_series')

    space_fav_season = parser.add_argument_group()
    space_fav_season.title = 'space_fav_season'
    space_fav_season.description = '获取合集或视频列表内视频'
    space_fav_season.add_argument('--by-space_fav_season', type=str,
                                  help='合集或视频列表 sid (或 URL)', dest='by_space_fav_season', default=None)

    favour_group = parser.add_argument_group()
    favour_group.title = 'favour'
    favour_group.description = '收藏夹'
    favour_group.add_argument(
        '--by-fav', type=str, help='收藏夹 fid (或 URL)', dest='by_fav', default=None)

    args = parser.parse_args()
    return args
 """


async def by_series(url_or_sid: str) -> Path:
    sid = sid = (
        re.search(r"sid=(\d+)", url_or_sid).groups()[0]
        if url_or_sid.startswith("http")
        else url_or_sid
    )  # type: ignore
    client = AsyncClient(**api.dft_client_settings)
    print(f"正在获取 {sid} 的视频列表……")
    col_name, up_name, bvids = await api.get_collect_info(client, sid)
    filepath = f"bvids/by-sapce_fav_season/sid-{sid}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        for bv_id in bvids:
            f.write(f"{bv_id}" + "\n")
    print(f"已获取 {col_name}（{up_name}）的 {len(bvids)} 个视频")
    print(f"到 {abs_filepath}")
    return Path(abs_filepath)


def by_ranking(rid: int) -> Path:
    bilibili_ranking_api = "https://api.bilibili.com/x/web-interface/ranking/v2"
    bilibili_ranking_params = {"rid": rid, "type": "all"}

    r = requests.get(bilibili_ranking_api, params=bilibili_ranking_params)
    r.raise_for_status()
    ranking_json = json.loads(r.text)
    assert ranking_json["code"] == 0  # 0 为成功（HTTP 200 不能信）

    ranking = ranking_json["data"]["list"]
    bvids = []
    for video_info in ranking:
        # print(video_info['title'], video_info['bvid'], video_info['pic'])
        bvid = video_info["bvid"]
        bvids.append(bvid)

    import datetime

    datetime.date.today()
    os.makedirs("bvids", exist_ok=True)

    bvids_filepath = f"bvids/by-ranking/rid-{rid}/rid-{rid}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(bvids_filepath), exist_ok=True)
    with open(bvids_filepath, "w", encoding="utf-8") as f:
        for bvid in bvids:
            f.write(f"{bvid}" + "\n")
    abs_filepath = os.path.abspath(bvids_filepath)
    print(f"已保存 {len(bvids)} 个 bvid 到 {abs_filepath}")
    return Path(abs_filepath)


async def by_up_videos(url_or_mid: str) -> Path:
    """频率高了会封"""

    if isinstance(url_or_mid, int):
        mid = str(url_or_mid)
    elif url_or_mid.startswith("http"):
        mid = re.findall(r"/(\d+)", url_or_mid)[0]
    else:
        mid = url_or_mid

    assert isinstance(mid, str)
    assert mid.isdigit(), "mid 应是数字字符串"

    client = AsyncClient(**api.dft_client_settings)
    ps = 30  # 每页视频数，最小 1，最大 50，默认 30
    order = "pubdate"  # 默认为pubdate 最新发布：pubdate 最多播放：click 最多收藏：stow
    keyword = ""  # 搜索关键词
    bv_ids = []
    pn = 1
    print(f"获取第 {pn} 页...")
    up_name, total_size, bv_ids_page = await api.get_up_info(
        client, mid, pn, ps, order, keyword
    )
    bv_ids += bv_ids_page
    print(
        f"{mid} {up_name} 共 {total_size} 个视频. (如果最新的视频为合作视频的非主作者，UP 名可能会识别错误，但不影响获取 bvid 列表)"
    )
    while pn < total_size / ps:
        pn += 1
        print(f"获取第 {pn} 页 (10s...)")
        await asyncio.sleep(10)
        _, _, bv_ids_page = await api.get_up_info(client, mid, pn, ps, order, keyword)
        bv_ids += bv_ids_page

    print(mid, up_name, total_size)
    await client.aclose()
    assert len(bv_ids) == len(set(bv_ids)), "有重复的 bv_id"
    assert total_size == len(bv_ids), "视频总数不匹配"
    filepath = f"bvids/by-up_videos/mid-{mid}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        for bv_id in bv_ids:
            f.write(f"{bv_id}" + "\n")
    print(f"已保存 {len(bv_ids)} 个 bvid 到 {abs_filepath}")
    return Path(abs_filepath)


def by_popular_precious():
    API_URL = "https://api.bilibili.com/x/web-interface/popular/precious"
    r = requests.get(API_URL)
    r.raise_for_status()
    popular_precious_json = json.loads(r.text)
    assert popular_precious_json["code"] == 0
    bvids = []
    for video_info in popular_precious_json["data"]["list"]:
        bvid = video_info["bvid"]
        bvids.append(bvid)
    filepath = f"bvids/by-popular_precious/{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        f.write("\n".join(bvids))
    print(f"已保存 {len(bvids)} 个 bvid 到 {abs_filepath}")


def by_popular_series_one(number: int):
    API_URL = "https://api.bilibili.com/x/web-interface/popular/series/one"
    params = {"number": number}
    r = requests.get(API_URL, params=params)
    r.raise_for_status()
    popular_series_json = json.loads(r.text)
    assert popular_series_json["code"] == 0
    bvids = []
    for video_info in popular_series_json["data"]["list"]:
        bvid = video_info["bvid"]
        bvids.append(bvid)
    filepath = f"bvids/by-popular_series/s{number}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        f.write("\n".join(bvids))
    print(f"已保存 {len(bvids)} 个 bvid 到 {abs_filepath}")


def not_got_popular_series() -> list[int]:
    API_URL = "http://api.bilibili.com/x/web-interface/popular/series/list"
    got_series = []
    os.makedirs("bvids/by-popular_series", exist_ok=True)
    for filename in os.listdir("bvids/by-popular_series"):
        if filename.endswith(".txt"):
            # s{number}-{int(time.time())}.txt
            got_series.append(int(filename.split("-")[0][1:]))
    r = requests.get(API_URL)
    r.raise_for_status()
    popular_series_json = json.loads(r.text)
    assert popular_series_json["code"] == 0
    max_series_number = popular_series_json["data"]["list"][0]["number"]
    series_not_got = []
    for i in range(1, max_series_number + 1):
        if i not in got_series:
            series_not_got.append(i)
    return series_not_got


async def by_favlist(url_or_fid: str):
    if url_or_fid.startswith("http"):
        fid = re.findall(r"fid=(\d+)", url_or_fid)[0]
    else:
        fid = url_or_fid

    client = AsyncClient(**api.dft_client_settings)
    PAGE_SIZE = 20
    media_left = None
    total_size = None
    bvids = []
    page_num = 1
    while media_left is None or media_left > 0:
        # bilix 的收藏夹获取有 bug
        fav_name, up_name, total_size, available_bvids = await api.get_favour_page_info(
            client=client, url_or_fid=fid, pn=page_num, ps=PAGE_SIZE, keyword=""
        )
        bvids.extend(available_bvids)
        if media_left is None:
            print(f"fav_name: {fav_name}, up_name: {up_name}, total_size: {total_size}")
        media_left = total_size - PAGE_SIZE * page_num
        print(f"还剩 ~{media_left // PAGE_SIZE} 页", end="\r")
        await asyncio.sleep(2)
        page_num += 1
    await client.aclose()
    assert total_size is not None
    assert len(bvids) == len(set(bvids)), "有重复的 bvid"
    print(f"{len(bvids)} 个有效视频，{total_size-len(bvids)} 个失效视频")
    filepath = f"bvids/by-favour/fid-{fid}-{int(time.time())}.txt"
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, "w", encoding="utf-8") as f:
        f.write("\n".join(bvids))
        f.write("\n")
    print(f"已保存 {len(bvids)} 个 bvid 到 {abs_filepath}")


async def main(
    series: str,
    ranking: str,
    rid: str,
    up_videos: str,
    popular_precious: bool,
    popular_series: bool,
    popular_series_number: int,
    all_popular_series: bool,
    favlist: str,
):
    if ranking:
        by_ranking(rid)
    if up_videos:
        await by_up_videos(up_videos)
    if popular_precious:
        by_popular_precious()
    if popular_series:
        if all_popular_series:
            for number in not_got_popular_series():
                time.sleep(3)
                by_popular_series_one(number)
        else:
            by_popular_series_one(popular_series_number)
    if series:
        await by_series(series)
    if favlist:
        await by_favlist(favlist)


class URLorIntParamType(click.ParamType):
    def __init__(self, name):
        self.name = "URL|" + name

    def convert(self, value, param, ctx):
        # Simple regex to check if value might be a URL
        # (just checking if it starts with http:// or https://)
        url_pattern = re.compile(r"^https?://")

        # If value matches URL pattern or is a digit, return value
        if url_pattern.match(value) or value.isdigit():
            return value

        # If value doesn't match any, raise an error
        self.fail(f"{value!r} is not a valid {self.name}", param, ctx)


@click.command(
    short_help=click.style("批量获取 BV 号", fg="cyan"),
    help="请通过 flag 指定至少一种批量获取 BV 号的方式。多个不同组的 flag 同时使用时，将会先后通过不同方式获取。",
)
@optgroup.group("合集")
@optgroup.option(
    "--series",
    "-s",
    help=click.style("合集或视频列表内视频", fg="red"),
    type=URLorIntParamType("sid"),
)
@optgroup.group("排行榜")
@optgroup.option(
    "--ranking",
    "-r",
    help=click.style("排行榜（全站榜，非个性推荐榜）", fg="yellow"),
    is_flag=True,
)
@optgroup.option(
    "--rid",
    "--ranking-id",
    default=0,
    show_default=True,
    help=click.style("目标排行 rid，0 为全站排行榜。rid 等于分区的 tid", fg="yellow"),
    type=int,
)
@optgroup.group("UP 主")
@optgroup.option(
    "--up-videos",
    "-u",
    help=click.style("UP 主用户页投稿", fg="cyan"),
    type=URLorIntParamType("mid"),
)
@optgroup.group("入站必刷")
@optgroup.option(
    "--popular-precious", help=click.style("入站必刷，更新频率低", fg="bright_red"), is_flag=True
)
@optgroup.group("每周必看")
@optgroup.option(
    "--popular-series",
    "-p",
    help=click.style("每周必看，每周五晚 18:00 更新", fg="magenta"),
    is_flag=True,
)
@optgroup.option(
    "--popular-series-number",
    default=1,
    type=int,
    show_default=True,
    help=click.style("获取第几期（周）", fg="magenta"),
)
@optgroup.option(
    "--all-popular-series",
    help=click.style("自动获取全部的每周必看（增量）", fg="magenta"),
    is_flag=True,
)
@optgroup.group("收藏夹")
@optgroup.option(
    "--favlist",
    "--fav",
    help=click.style("收藏夹", fg="green"),
    type=URLorIntParamType("fid"),
)
def get(**kwargs):
    if (
        not kwargs["favlist"]
        and not kwargs["series"]
        and not kwargs["ranking"]
        and not kwargs["up_videos"]
        and not kwargs["popular_precious"]
        and not kwargs["popular_series"]
    ):
        click.echo(click.style("ERROR: 请指定至少一种获取方式。", fg="red"))
        click.echo(get.get_help(click.Context(get)))
        return
    asyncio.run(main(**kwargs))
